#!/usr/bin/python3
"""
## © University of Southampton IT Innovation Centre, 2018
##
## Copyright in this software belongs to University of Southampton
## IT Innovation Centre of Gamma House, Enterprise Road,
## Chilworth Science Park, Southampton, SO16 7NS, UK.
##
## This software may not be used, sold, licensed, transferred, copied
## or reproduced in whole or in part in any manner or form or in or
## on any media by any person other than in accordance with the terms
## of the Licence Agreement supplied with the software, or otherwise
## without the prior written consent of the copyright owners.
##
## This software is distributed WITHOUT ANY WARRANTY, without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
## PURPOSE, except where stated in the Licence Agreement supplied with
## the software.
##
##      Created By :            Nikolay Stanchev
##      Created Date :          17-04-2018
##      Created for Project :   FLAME
"""

import pytest
import random
import time
import threading


class TestE2ESimulation(object):
    """
    A testing class used to group all the tests related to the E2E simulation data
    """

    @pytest.fixture(scope='class', autouse=True)
    def run_simulator(self, e2e_simulator, e2e_aggregator):
        """
        A fixture, which runs the simulation before running the tests.

        :param e2e_simulator: the simulator for the end-to-end data
        :param e2e_aggregator: the aggregator which merges the network and service measurements
        """

        random.seed(0)  # Seed random function so we can reliably test for average queries

        print("Starting aggregator...")
        event = threading.Event()
        e2e_aggregator.set_event_lock(event)
        e2e_aggregator.start()

        event.wait()  # wait until the aggregator thread has set the event lock (it has reached its run method and is ready to start)

        print("Running simulation, please wait...")
        e2e_simulator.run()

        print("Waiting for INFLUX to finish receiving simulation data...")
        time.sleep(e2e_simulator.SIMULATION_LENGTH)  # wait for data to finish arriving at the INFLUX database
        print("... simulation data fixture finished")

        print("... stopping aggregator")
        e2e_aggregator.stop()

    @pytest.mark.parametrize("query, expected_result", [
        ('SELECT count(*) FROM "E2EMetrics"."autogen"."network_delays"',
         {"time": "1970-01-01T00:00:00Z", "count_latency": 120, "count_bandwidth": 120}),
        ('SELECT count(*) FROM "E2EMetrics"."autogen"."service_delays"',
         {"time": "1970-01-01T00:00:00Z", "count_response_time": 24, "count_request_size": 24, "count_response_size": 24}),
        ('SELECT count(*) FROM "E2EMetrics"."autogen"."e2e_delays"',
         {"time": "1970-01-01T00:00:00Z", "count_delay_forward": 38, "count_delay_reverse": 38, "count_delay_service": 38,
          "count_avg_request_size": 38, "count_avg_response_size": 38, "count_avg_bandwidth": 38}),

        ('SELECT mean(*) FROM "E2EMetrics"."autogen"."e2e_delays"',
         {"time": "1970-01-01T00:00:00Z", "mean_delay_forward": 8.010964912280702, "mean_delay_reverse": 12.881578947368423, "mean_delay_service": 23.42105263157895,
          'mean_avg_request_size': 10485760, 'mean_avg_response_size': 1024, 'mean_avg_bandwidth': 104857600}),
        ])
    def test_simulation(self, influx_db, query, expected_result):
        """
        This is the entry point of the test. This method will be found and executed when the module is ran using pytest

        :param query: the query to execute (value obtained from the pytest parameter decorator)
        :param expected_result: the result expected from executing the query (value obtained from the pytest parameter decorator)
        :param influx_db the import db client fixture - imported from contest.py
        """

        # pytest automatically goes through all queries under test, declared in the parameters decorator
        print("\n")  # prints a blank line for formatting purposes

        # the raise_errors=False argument is given so that we could actually test that the DB didn't return any errors instead of raising an exception
        query_result = influx_db.query(query, raise_errors=False)

        # test the error attribute of the result is None, that is no error is returned from executing the DB query
        assert query_result.error is None, "An error was encountered while executing query {0}.".format(query)

        # get the dictionary of result points; the next() function just gets the first element of the query results generator (we only expect one item in the generator)
        actual_result = next(query_result.get_points())

        assert expected_result == actual_result, "E2E Simulation test failure"

        print("Successfully passed test for the following query: {0}".format(query))
